#!/usr/bin/gawk -f

BEGIN {
    SCRIPTNAME="getverdiff.awk"
    if (ARGV[1]=="-h" || ARGV[1]=="--help") {
        print "Usage: " SCRIPTNAME " [OPTIONS] <getver.log> ...  "
        print "Usage: " SCRIPTNAME " [OPTIONS] < input  "
        print "Compare getver log files"
        print "OPTIONS"
        print "-v MESSAGE_LANG ja|en   Output language, either ja or en"
        print "-v SEP <char>           tab character"
        print "-v QUOTE <char>         quotation character"
        print "-v TIMEZONE <zone>      check that all files match given timezone"
        exit;
    }
    FILELISTING_INDEX_COUNTER = 1;
    FILELISTING_INFO_INDEX_COUNTER = 1;
    TIMEZONE = "";
    ERROR=0;
    STARTLETTER = 65;

    if (ARGC<3) {
        FILECOUNT = -1
    }
    if (SEP=="") {
        SEP = ",";
    }
    if (QUOTE=="") {
        QUOTE = "\"";
    }
    if (MESSAGE_LANG=="") {
        MESSAGE_LANG="ja";
    }

    # tracking message
    MESSAGE["en","r0"]=":renamed to  :";
    MESSAGE["en","r1"]=":renamed from:";
    MESSAGE["en","m0"]=":moved to    :";
    MESSAGE["en","m1"]=":moved from  :";
    MESSAGE["en","c0"]=":copies in   :";
    MESSAGE["en","c1"]=":copies from :";

    # 2-file comparison results 
    MESSAGE["en","00"]="invalid";
    MESSAGE["en","0A"]="added in MATCH";
    MESSAGE["en","0B"]="invalid";
    MESSAGE["en","A0"]="deleted in MATCH";
    MESSAGE["en","AA"]="identical";
    MESSAGE["en","AB"]="regressed in MATCH";
    MESSAGE["en","B0"]="invalid";
    MESSAGE["en","BA"]="updated in MATCH";
    MESSAGE["en","BB"]="invalid";

    # 3-file comparison results 
    MESSAGE["en","000"]="invalid";
    MESSAGE["en","00A"]="added only in M2";
    MESSAGE["en","00B"]="invalid";
    MESSAGE["en","00C"]="invalid";
    MESSAGE["en","0A0"]="identically deleted";
    MESSAGE["en","0AA"]="deleted only in M1";
    MESSAGE["en","0AB"]="deleted in M1 but regressed in M2";
    MESSAGE["en","0AC"]="invalid";
    MESSAGE["en","0B0"]="invalid";
    MESSAGE["en","0BA"]="deleted in M1 but updated in M2";
    MESSAGE["en","0BB"]="invalid";
    MESSAGE["en","0BC"]="invalid";
    MESSAGE["en","0C0"]="invalid";
    MESSAGE["en","0CA"]="invalid";
    MESSAGE["en","0CB"]="invalid";
    MESSAGE["en","0CC"]="invalid";
    MESSAGE["en","A00"]="added only in M1";
    MESSAGE["en","A0A"]="identically added";
    MESSAGE["en","A0B"]="conflicting additions in M1 and M2";
    MESSAGE["en","A0C"]="invalid";
    MESSAGE["en","AA0"]="deleted only in M2";
    MESSAGE["en","AAA"]="identical";
    MESSAGE["en","AAB"]="regressed only in M2";
    MESSAGE["en","AAC"]="invalid";
    MESSAGE["en","AB0"]="updated in M1 but deleted in M2";
    MESSAGE["en","ABA"]="identically updated";
    MESSAGE["en","ABB"]="updated only in M1";
    MESSAGE["en","ABC"]="updated in M1 but regressed in M2";
    MESSAGE["en","AC0"]="invalid";
    MESSAGE["en","ACA"]="invalid";
    MESSAGE["en","ACB"]="conflicting updates in M1 and M2";
    MESSAGE["en","ACC"]="invalid";
    MESSAGE["en","B00"]="invalid";
    MESSAGE["en","B0A"]="conflicting additions in M2 and M1";
    MESSAGE["en","B0B"]="invalid";
    MESSAGE["en","B0C"]="invalid";
    MESSAGE["en","BA0"]="regressed in M1 but deleted in M2";
    MESSAGE["en","BAA"]="regressed only in M1";
    MESSAGE["en","BAB"]="identically regressed ";
    MESSAGE["en","BAC"]="conflicting regressions in M1 and M2";
    MESSAGE["en","BB0"]="invalid";
    MESSAGE["en","BBA"]="updated only in M2";
    MESSAGE["en","BBB"]="invalid";
    MESSAGE["en","BBC"]="invalid";
    MESSAGE["en","BC0"]="invalid";
    MESSAGE["en","BCA"]="conflicting updates in M2 and M1";
    MESSAGE["en","BCB"]="invalid";
    MESSAGE["en","BCC"]="invalid";
    MESSAGE["en","C00"]="invalid";
    MESSAGE["en","C0A"]="invalid";
    MESSAGE["en","C0B"]="invalid";
    MESSAGE["en","C0C"]="invalid";
    MESSAGE["en","CA0"]="invalid";
    MESSAGE["en","CAA"]="invalid";
    MESSAGE["en","CAB"]="conflicting regressions in M2 and M1";
    MESSAGE["en","CAC"]="invalid";
    MESSAGE["en","CB0"]="invalid";
    MESSAGE["en","CBA"]="regressed in M1 but updated in M2";
    MESSAGE["en","CBB"]="invalid";
    MESSAGE["en","CBC"]="invalid";
    MESSAGE["en","CC0"]="invalid";
    MESSAGE["en","CCA"]="invalid";
    MESSAGE["en","CCB"]="invalid";
    MESSAGE["en","CCC"]="invalid";

    # tracking message
    MESSAGE["ja", "r0"]="";
    MESSAGE["ja", "r1"]="";
    MESSAGE["ja", "m0"]="ړ";
    MESSAGE["ja", "m1"]="ړ";
    MESSAGE["ja", "c0"]="ʖ";
    MESSAGE["ja", "c1"]="ʖ";

    # 2-file comparison results 
    MESSAGE["ja","00"]="invalid";
    MESSAGE["ja","0A"]="MATCHɂĒǉ";
    MESSAGE["ja","0B"]="invalid";
    MESSAGE["ja","A0"]="MATCHɂč폜";
    MESSAGE["ja","AA"]="v";
    MESSAGE["ja","AB"]="MATCHɂĕ";
    MESSAGE["ja","B0"]="invalid";
    MESSAGE["ja","BA"]="MATCHɂčXV";
    MESSAGE["ja","BB"]="invalid";

    # 3-file comparison results 
    MESSAGE["ja","000"]="invalid";
    MESSAGE["ja","00A"]="M2̂ݒǉ";
    MESSAGE["ja","00B"]="invalid";
    MESSAGE["ja","00C"]="invalid";
    MESSAGE["ja","0A0"]="v폜";
    MESSAGE["ja","0AA"]="M1̂ݍ폜";
    MESSAGE["ja","0AB"]="M1ɂč폜;M2ɂĕ";
    MESSAGE["ja","0AC"]="invalid";
    MESSAGE["ja","0B0"]="invalid";
    MESSAGE["ja","0BA"]="M1ɂč폜;M2ɂčXV";
    MESSAGE["ja","0BB"]="invalid";
    MESSAGE["ja","0BC"]="invalid";
    MESSAGE["ja","0C0"]="invalid";
    MESSAGE["ja","0CA"]="invalid";
    MESSAGE["ja","0CB"]="invalid";
    MESSAGE["ja","0CC"]="invalid";
    MESSAGE["ja","A00"]="M1̂ݒǉ";
    MESSAGE["ja","A0A"]="vǉ";
    MESSAGE["ja","A0B"]="ǉ:M1ŐV";
    MESSAGE["ja","A0C"]="invalid";
    MESSAGE["ja","AA0"]="M2̂ݍ폜";
    MESSAGE["ja","AAA"]="v";
    MESSAGE["ja","AAB"]="M2̂ݕ";
    MESSAGE["ja","AAC"]="invalid";
    MESSAGE["ja","AB0"]="M1ɂčXV;M2ɂč폜";
    MESSAGE["ja","ABA"]="vXV";
    MESSAGE["ja","ABB"]="M1̂ݍXV";
    MESSAGE["ja","ABC"]="M1ɂčXV;M2ɂĕ";
    MESSAGE["ja","AC0"]="invalid";
    MESSAGE["ja","ACA"]="invalid";
    MESSAGE["ja","ACB"]="XV:M1ŐV";
    MESSAGE["ja","ACC"]="invalid";
    MESSAGE["ja","B00"]="invalid";
    MESSAGE["ja","B0A"]="ǉ:M2ŐV";
    MESSAGE["ja","B0B"]="invalid";
    MESSAGE["ja","B0C"]="invalid";
    MESSAGE["ja","BA0"]="M1ɂĕ;M2ɂč폜";
    MESSAGE["ja","BAA"]="M1̂ݕ";
    MESSAGE["ja","BAB"]="v";
    MESSAGE["ja","BAC"]=":M1ŐV";
    MESSAGE["ja","BB0"]="invalid";
    MESSAGE["ja","BBA"]="M2̂ݍXV";
    MESSAGE["ja","BBB"]="invalid";
    MESSAGE["ja","BBC"]="invalid";
    MESSAGE["ja","BC0"]="invalid";
    MESSAGE["ja","BCA"]="XV:M2ŐV";
    MESSAGE["ja","BCB"]="invalid";
    MESSAGE["ja","BCC"]="invalid";
    MESSAGE["ja","C00"]="invalid";
    MESSAGE["ja","C0A"]="invalid";
    MESSAGE["ja","C0B"]="invalid";
    MESSAGE["ja","C0C"]="invalid";
    MESSAGE["ja","CA0"]="invalid";
    MESSAGE["ja","CAA"]="invalid";
    MESSAGE["ja","CAB"]=":M2ŐV";
    MESSAGE["ja","CAC"]="invalid";
    MESSAGE["ja","CB0"]="invalid";
    MESSAGE["ja","CBA"]="M1ɂĕ;M2ɂčXV";
    MESSAGE["ja","CBB"]="invalid";
    MESSAGE["ja","CBC"]="invalid";
    MESSAGE["ja","CC0"]="invalid";
    MESSAGE["ja","CCA"]="invalid";
    MESSAGE["ja","CCB"]="invalid";
    MESSAGE["ja","CCC"]="invalid";
}

PREVFILE!=FILENAME {
    FILECOUNT = FILECOUNT + 1;
    FILENAMES[FILECOUNT] = FILENAME;
}


/^\s*$/ {
    next;
}

/\/$/{
    next;
}

{
    #SAMPLE LINE
    #19 0,0,0,0 2009-08-27 15:36:24@+9:00 Additional Info dd478a727ae178b1a7f2a1b3d745d3f3f0e2f7a9 *./obj/vfstab
    match($0, "^([0-9]+ |)([^ ]+ |)([0-9][0-9][0-9][0-9]-[0-9][0-9]-[0-9][0-9] |)([0-9][0-9]:[0-9][0-9]:[0-9][0-9]|)(@[+-][0-9]+:[0-9][0-9] |)(.*?|)\\*(.*)", ENTRY);
    match(ENTRY[6], "(.*)([[:alnum:]][[:alnum:]][[:alnum:]][[:alnum:]][[:alnum:]][[:alnum:]][[:alnum:]][[:alnum:]][[:alnum:]][[:alnum:]][[:alnum:]][[:alnum:]][[:alnum:]][[:alnum:]][[:alnum:]][[:alnum:]][[:alnum:]][[:alnum:]][[:alnum:]][[:alnum:]][[:alnum:]][[:alnum:]][[:alnum:]][[:alnum:]][[:alnum:]][[:alnum:]][[:alnum:]][[:alnum:]][[:alnum:]][[:alnum:]][[:alnum:]][[:alnum:]][[:alnum:]][[:alnum:]][[:alnum:]][[:alnum:]][[:alnum:]][[:alnum:]][[:alnum:]][[:alnum:]])[[:space:]]*$", ENTRYINFO);

    ENTRYNAME = ENTRY[7];
    ENTRYNAME = gensub("\\\\", "/", "g", ENTRYNAME);
    ENTRYNAME = gensub("^[^/]*[\\/]", "./", "", ENTRYNAME);
    ENTRYNAME = gensub("[/]", "\0", "g", ENTRYNAME);

    ENTRYROOT = ENTRY[7];
    ENTRYROOT = gensub("\\\\", "/", "g", ENTRYROOT);
    ENTRYROOT = gensub("[/].*", "", "", ENTRYROOT);

    if (length(ENTRYNAME) > MAX_ENTRYNAME_LEN) {
        MAX_ENTRYNAME_LEN = length(ENTRYNAME);
    }

}

(ARGC<3) && (ENTRYROOT!=PREVROOT) {
    FILECOUNT = FILECOUNT + 1;
    FILENAMES[FILECOUNT] = ENTRYROOT;
}

{

    IDX = FILELISTING_INDEX[ENTRYNAME];
    if (IDX=="") {
        IDX = FILELISTING_INDEX_COUNTER++;
        FILELISTING_INDEX[ENTRYNAME] = IDX;
        FILELISTING_NAMES[IDX] = ENTRYNAME;
    }

    FILELISTING_INFO[IDX,FILECOUNT,"SIZE"] = ENTRY[1];
    FILELISTING_INFO[IDX,FILECOUNT,"VER"] = ENTRY[2];
    FILELISTING_INFO[IDX,FILECOUNT,"MOD_DATE"] = ENTRY[3];
    FILELISTING_INFO[IDX,FILECOUNT,"MOD_TIME"] = ENTRY[4];
    FILELISTING_INFO[IDX,FILECOUNT,"MOD_TIMEZONE"] = gensub("@","", "g", ENTRY[5]);

    MOD_TIMEZONE_HOUR = gensub(":.*", "", "", FILELISTING_INFO[IDX,FILECOUNT,"MOD_TIMEZONE"]);
    if (TIMEZONE == "" ) {
        MOD_TIMEZONE_COMMON = MOD_TIMEZONE_HOUR
    } else {
        MOD_TIMEZONE_COMMON = gensub(":.*", "", "", TIMEZONE);
    }
    match(FILELISTING_INFO[IDX,FILECOUNT,"MOD_TIME"], "([0-9][0-9]):([0-9][0-9]):([0-9][0-9])", MOD_TIME);
    MOD_TIME[1] = MOD_TIME[1] - MOD_TIMEZONE_HOUR + MOD_TIMEZONE_COMMON
    if (MOD_TIME[1] < 0) {
        MOD_TIME[1]=MOD_TIME[1]+24
        FILELISTING_INFO[IDX,FILECOUNT,"MOD_DATE"] = calendarday(julianday(FILELISTING_INFO[IDX,FILECOUNT,"MOD_DATE"]) - 1) " ";
    }
    if (MOD_TIME[1] > 23) {
        MOD_TIME[1]=MOD_TIME[1]-24
        FILELISTING_INFO[IDX,FILECOUNT,"MOD_DATE"] = calendarday(julianday(FILELISTING_INFO[IDX,FILECOUNT,"MOD_DATE"]) + 1) " ";
    }
    FILELISTING_INFO[IDX,FILECOUNT,"MOD_TIME"] = sprintf("%02d:%02d:%02d", MOD_TIME[1], MOD_TIME[2], MOD_TIME[3])
    FILELISTING_INFO[IDX,FILECOUNT,"MOD_TIMEZONE"] = sprintf("%+0d:00", MOD_TIMEZONE_COMMON)
    
    FILELISTING_INFO[IDX,FILECOUNT,"MOD_TIMESTAMP"] = FILELISTING_INFO[IDX,FILECOUNT,"MOD_DATE"] " " FILELISTING_INFO[IDX,FILECOUNT,"MOD_TIME"];

    FILELISTING_INFO[IDX,FILECOUNT,"INFO"] = ENTRY[6];
    if (""!=ENTRYINFO[2]) {
        FILELISTING_INFO[IDX,FILECOUNT,"CHECKSUM"] = ENTRYINFO[2];
    } else {
        FILELISTING_INFO[IDX,FILECOUNT,"CHECKSUM"] = FILELISTING_INFO[IDX,FILECOUNT,"SIZE"] " " FILELISTING_INFO[IDX,FILECOUNT,"MOD_DATE"] " " FILELISTING_INFO[IDX,FILECOUNT,"MOD_TIME"] " " FILELISTING_INFO[IDX,FILECOUNT,"MOD_TIMEZONE"]
    }

    TRACKERENTRY = gensub("^[^/]*[\\/]", "./", "", ENTRY[7]);
    FILEPATHONLY = gensub(/\/[^\/]*$/,"/","g",TRACKERENTRY);
    FILENAMEONLY = gensub(/^.*\//,"","g",TRACKERENTRY);
    FILELISTING_TRACKER[FILECOUNT,FILELISTING_INFO[IDX,FILECOUNT,"CHECKSUM"],FILENAMEONLY]=FILELISTING_TRACKER[FILECOUNT,FILELISTING_INFO[IDX,FILECOUNT,"CHECKSUM"],FILENAMEONLY] "\0" FILEPATHONLY;
    FILELISTING_TRACKER[FILECOUNT,FILELISTING_INFO[IDX,FILECOUNT,"CHECKSUM"],FILEPATHONLY]=FILELISTING_TRACKER[FILECOUNT,FILELISTING_INFO[IDX,FILECOUNT,"CHECKSUM"],FILEPATHONLY] "\0" FILENAMEONLY;
    # print "FILECOUNT=" FILECOUNT 
    # print "-IDX=" IDX 
    # print "-FILEPATHONLY=" FILEPATHONLY 
    # print "-FILENAMEONLY=" FILENAMEONLY 
    # print "-PATHMATCH=" FILELISTING_TRACKER[FILECOUNT,FILELISTING_INFO[IDX,FILECOUNT,"CHECKSUM"],FILENAMEONLY]
    # print "-NAMEMATCH=" FILELISTING_TRACKER[FILECOUNT,FILELISTING_INFO[IDX,FILECOUNT,"CHECKSUM"],FILEPATHONLY]
    # print "-MATCHKEY=" FILECOUNT "," FILELISTING_INFO[IDX,FILECOUNT,"CHECKSUM"]

    if (TIMEZONE=="") { 
        TIMEZONE = FILELISTING_INFO[IDX,FILECOUNT,"MOD_TIMEZONE"];
    } else {
        if (""!=FILELISTING_INFO[IDX,FILECOUNT,"MOD_TIMEZONE"]) {
            if (TIMEZONE != FILELISTING_INFO[IDX,FILECOUNT,"MOD_TIMEZONE"]) {
                print "error: Timezone mismatch in line "FNR" in " FILENAME > "/dev/stderr";
                ERROR=1;
                exit 1;
            }
        }
    }

    CONTENT = FILELISTING_INFO[IDX,FILECOUNT,"SIZE"]" "FILELISTING_INFO[IDX,FILECOUNT,"SIZE"] " " FILELISTING_INFO[IDX,FILECOUNT,"MOD_DATE"] " " FILELISTING_INFO[IDX,FILECOUNT,"MOD_TIME"] " " FILELISTING_INFO[IDX,FILECOUNT,"MOD_TIMEZONE"] " " FILELISTING_INFO[IDX,FILECOUNT,"CHECKSUM"]
    CONTIDX = FILELISTING_INFO_INDEX[CONTENT];
    if (CONTIDX=="") {
        CONTIDX = FILELISTING_INFO_INDEX_COUNTER++;
        FILELISTING_INFO_INDEX[CONTENT] = CONTIDX;
    }
    FILELISTING_INFO[IDX,FILECOUNT,"CONTENT"] = FILELISTING_INFO_INDEX[CONTENT];
}

{
    PREVFILE = FILENAME
    PREVROOT = ENTRYROOT
}

END {
    for (i in MESSAGE) {
        if (match(i, "^"MESSAGE_LANG)) {
            if (length(MESSAGE[i]) > maxlen) {
                maxlen = length(MESSAGE[i]);
            }
        }
    }
    for (i in MESSAGE) {
        if (match(i, "^"MESSAGE_LANG)) {
            while (length(MESSAGE[i]) < maxlen) {
                MESSAGE[i] = MESSAGE[i] " ";
            }
        }
    }

    if (ARGV[1]=="-h" || ARGV[1]=="--help") {
        exit;
    }
    if (ERROR) {
        exit 1;
    }
    if (FILECOUNT==2) {
        print sprintf("%-"(MAX_ENTRYNAME_LEN+3)"s", " ") SEP "SOURCE     " SEP "MATCH      "
        print "# SOURCE  =" FILENAMES[1];
        print "# MATCH   =" FILENAMES[2];
    }
    if (FILECOUNT==3) {
        print sprintf("%-"(MAX_ENTRYNAME_LEN+3)"s", " ") SEP "MATCH-M1   " SEP "SOURCE     " SEP "MATCH-M2   "
        print "# MATCH-M1=" FILENAMES[1];
        print "# SOURCE  =" FILENAMES[2];
        print "# MATCH-M2=" FILENAMES[3];
    }
    if (FILECOUNT>3) {
        BUFFER = sprintf("%-"(MAX_ENTRYNAME_LEN+3)"s", " ") SEP 
        for (j=1; j<=FILECOUNT;j++) {
            BUFFER = BUFFER "SOURCE-S" j " " SEP
        }
        print BUFFER
        for (j=1; j<=FILECOUNT;j++) {
            print "# SOURCE-S"j"=" FILENAMES[j];
        }
    }

    asort(FILELISTING_NAMES, FILELISTING_NAMES_SORTED);
    for (i=1;i<=length(FILELISTING_INDEX);i++) {
        ENTRYNAME = FILELISTING_NAMES_SORTED[i];
        FILEIDX = FILELISTING_INDEX[ENTRYNAME];
        ENTRYNAME = QUOTE gensub(QUOTE, QUOTE QUOTE, "g", gensub("\0", "/", "g", ENTRYNAME)) QUOTE;
        BUFFER = sprintf("%-"(MAX_ENTRYNAME_LEN+3)"s", ENTRYNAME) SEP;
        for (j=1; j<=FILECOUNT;j++) {
            #ITEM = FILELISTING_INFO[FILEIDX,j,"CONTENT"];
            ITEM = FILELISTING_INFO[FILEIDX,j,"MOD_DATE"];
            if (ITEM!="") {
                BUFFER = BUFFER ITEM SEP;
            } else {
                BUFFER = BUFFER "0          " SEP;
            }
            TIMESTAMP[j]=FILELISTING_INFO[FILEIDX,j,"MOD_TIMESTAMP"] "\0" j;
        }
        asort(TIMESTAMP, TIMESTAMP_IDX);
        CODELETTER = STARTLETTER;
        delete CODE_IDX;

        SOURCEFILE=1;
        if (FILECOUNT==3) {
            SOURCEFILE=2;
        }
        for (j=FILECOUNT; j>=1;j--) {

            # Assign letters A,B,C... for unique content,
            # in order of file modification, with "A"
            # assigned to the newest content.
            TIMEIDX=substr(TIMESTAMP_IDX[j], index(TIMESTAMP_IDX[j], "\0")+1);
            CHK=FILELISTING_INFO[FILEIDX,TIMEIDX,"CHECKSUM"];

            REFERENCETARGET[TIMEIDX] = "";

            SRC = "S0"
            TGT = "S" TIMEIDX
            if (FILECOUNT==2) {
                SRC = "SOURCE"
                TGT = "MATCH"
            }
            if (FILECOUNT==3) {
                SRC ="SOURCE" 
                if (TIMEIDX==1) { TGT="M1" }
                if (TIMEIDX==3) { TGT="M2" }
            }

            if (""==FILELISTING_INFO[FILEIDX,TIMEIDX,"SIZE"]) {

                # If file does not exist in this
                # listing, assign "0" instead
                CODE[TIMEIDX] = 48;

                # set "d" for deleted
                TRACK[TIMEIDX] = 100;

                # MATCH has been deleted but let's try to find where it has gone.
                if (TIMEIDX!=SOURCEFILE) {

                    FILEPATHONLY = gensub("^\"", "", "", gensub(/\/[^\/]*$/,"/","g",ENTRYNAME));
                    FILENAMEONLY = gensub("\"$", "", "", gensub(/^.*\//,"","g",ENTRYNAME));

                    # to detected that the file has been moved away, we look for matches in the MATCH tree
                    LOCAL_PATH_MATCHES = FILELISTING_TRACKER[TIMEIDX,FILELISTING_INFO[FILEIDX,SOURCEFILE,"CHECKSUM"],FILENAMEONLY];
                    LOCAL_NAME_MATCHES = FILELISTING_TRACKER[TIMEIDX,FILELISTING_INFO[FILEIDX,SOURCEFILE,"CHECKSUM"],FILEPATHONLY];

                    # to confirm it was a move we need to chack that the SOURCE tree did not already have that file in the first place
                    PATH_MATCHES = FILELISTING_TRACKER[SOURCEFILE,FILELISTING_INFO[FILEIDX,SOURCEFILE,"CHECKSUM"],FILENAMEONLY];
                    NAME_MATCHES = FILELISTING_TRACKER[SOURCEFILE,FILELISTING_INFO[FILEIDX,SOURCEFILE,"CHECKSUM"],FILEPATHONLY];

                    MATCHED=0

                    # Detect file rename for deleted file
                    # To detect that file has has been renamed in MATCH, we need to find a file matching the coditions:
                    # 1. checksum matches
                    # 2. a) filename is either a substring or a superstring, excluding extension, 
                    #    b) but not matching exacatly
                    # 3. exists in the same directory
                    # 4. SOURCE did not already contain the file 
                    # 5. only one such file satisfying the above exists
                    if ((MATCHED==0) && (""!=LOCAL_NAME_MATCHES) && ("\0" FILENAMEONLY !=LOCAL_NAME_MATCHES)) {

                        MATCHED = 0;

                        # conditions (1) and (3) matched

                        MATCHCOUNT = split(LOCAL_NAME_MATCHES, MATCHINGFILES, "\0");

                        for (m=1; m<=length(MATCHINGFILES); m++) {
                            if (""==MATCHINGFILES[m]) {
                                continue;
                            }

                            CANDIDATE_ENTRY = MATCHINGFILES[m];
                            CANDIDATE_TOKEN = gensub("\\..*","","",MATCHINGFILES[m]);
                            FILENAME_TOKEN = gensub("\\..*","","",FILENAMEONLY);

                            COPIES = "";

                            if ((""!=CANDIDATE_TOKEN)&&(CANDIDATE_ENTRY!=FILENAMEONLY)&&((0!=index(CANDIDATE_TOKEN, FILENAME_TOKEN)) || (0!=index(FILENAME_TOKEN, CANDIDATE_TOKEN)))) {

                                # condition (2) matched
                                # now we have suspects which are likely related to the file, which we interprete as copies.

                                # set "c" for copy
                                TRACK[TIMEIDX] = 99;

#print "r0" CANDIDATE_ENTRY
                                nameof_sourcefile = gensub("[/]", "\0", "g",gensub("^[^/]*[\\/]", "./", "",gensub("\\\\", "/", "g", FILEPATHONLY CANDIDATE_ENTRY)));
                                sizeof_sourcefile = FILELISTING_INFO[FILELISTING_INDEX[nameof_sourcefile],SOURCEFILE,"SIZE"]

                                if (""==sizeof_sourcefile) {
                                    # condition (4)  satisfied

                                    if (MATCHED==0) {

                                        # RENAME
                                        # assume condition (5) has been met and treat it as a rename
                                        # set "r" for renamed

                                        TRACK[TIMEIDX] = 114;
                                        MATCHED=1;

                                        REFERENCETARGET[TIMEIDX]=REFERENCETARGET[TIMEIDX] "\0(" TGT ")" CANDIDATE_ENTRY

                                    } else {

                                        # condition (5) broken
                                        # downgread to "c" for copy

                                        TRACK[TIMEIDX] = 99;
                                    }
                                }
                                COPIES = COPIES "\0(" SRC ")" CANDIDATE_ENTRY
                            }
                        }
                        if (TRACK[TIMEIDX]==99) {
                            REFERENCETARGET[TIMEIDX]=COPIES
                        }
                    } 

                    # Detect file move for deleted file
                    # To detect that file has has been moved elsewhere in MATCH, the conditions are:
                    # 1. filename and checksum match 
                    # 2. exists in a different directory in SOURCE than in MATCH
                    # 3. and only one such file exists
                    # 4. SOURCE did not already contain the file found in (2)
                    if ((MATCH==0)&&(""!=LOCAL_PATH_MATCHES)&&("\0" FILEPATHONLY !=LOCAL_PATH_MATCHES)) {

                        MATCHED = 0;

                        # condition (1) passed
                        # now we know that copies of the file exist somewhere

                        # set "c" for copy
                        TRACK[TIMEIDX] = 99;

                        MATCHCOUNT = split(LOCAL_PATH_MATCHES, MATCHINGFILES, "\0");

                        for (m=1; m<=length(MATCHINGFILES); m++) {
                            if (""==MATCHINGFILES[m]) {
                                continue;
                            }

                            CANDIDATE_ENTRY = MATCHINGFILES[m];

                            # Since every entry starts with \0 there would be one extra entry at the beginning of the array
                            if (FILEPATHONLY!=CANDIDATE_ENTRY) {
                                # condition (2) and (3) pased

#print "m0" CANDIDATE_ENTRY
                                nameof_sourcefile = gensub("[/]", "\0", "g",gensub("^[^/]*[\\/]", "./", "",gensub("\\\\", "/", "g", CANDIDATE_ENTRY FILENAMEONLY)));
                                sizeof_sourcefile =  FILELISTING_INFO[FILELISTING_INDEX[nameof_sourcefile],SOURCEFILE,"SIZE"];

                                if(""==sizeof_sourcefile) {
                                    # condition (4) passed

                                    if (MATCHED==0) {

                                        # MOVE detected
                                        # assume condition (3) passed
                                        # set "m" for moved

                                        TRACK[TIMEIDX] = 109;
                                        MATCHED=1;

                                        REFERENCETARGET[TIMEIDX]="\0(" TGT ")" CANDIDATE_ENTRY
                                    } else {

                                        # condition (3) broken
                                        # downgread to "c" for copy

                                        TRACK[j] = 99;
                                    }
                                }
                            }
                        }
                        if (TRACK[TIMEIDX]==99) {
                            filepathpos = index(LOCAL_PATH_MATCHES "\0", "\0" FILEPATHONLY "\0");
                            if (filepathpos==0) {
                                REFERENCETARGET[TIMEIDX]="\0(" TGT ")" substr(LOCAL_PATH_MATCHES,2);
                            } else {
                                REFERENCETARGET[TIMEIDX]="\0(" TGT ")" substr(substr(LOCAL_PATH_MATCHES, 1, filepathpos-1) substr(LOCAL_PATH_MATCHES, filepathpos+length(FILEPATHONLY)+1), 2);
                            }
                        }
                    } 
                }

            } else {
                IDX=CODE_IDX[CHK]
                if (IDX=="") {
                    IDX=CODELETTER;
                    CODE_IDX[CHK]=IDX;
                    CODELETTER++;
                }
                CODE[TIMEIDX] = IDX;

                # set "e" for edited
                TRACK[TIMEIDX] = 101;
            }
            PREVCHK = CHK;
        }
        for (j=1; j<=FILECOUNT;j++) {
            SRC = "S0"
            TGT = "S" j
            if (FILECOUNT==2) {
                SRC = "SOURCE"
                TGT = "MATCH"
            }
            if (FILECOUNT==3) {
                SRC ="SOURCE" 
                if (j==1) { TGT="M1" }
                if (j==3) { TGT="M2" }
            }
            # If file did not exist in SOURCE, let's check if this
            # is a new additon or a copy of something else by
            # looking for the file elsewhere.
            if ((j!=SOURCEFILE) && (CODE[j]!=48) && CODE[SOURCEFILE]==48) {

                FILEPATHONLY = gensub("^\"", "", "", gensub(/\/[^\/]*$/,"/","g",ENTRYNAME));
                FILENAMEONLY = gensub("\"$", "", "", gensub(/^.*\//,"","g",ENTRYNAME));

                # to detect that the file was moved in, we check against the SOURCE tree
                PATH_MATCHES = FILELISTING_TRACKER[SOURCEFILE,FILELISTING_INFO[FILEIDX,j,"CHECKSUM"],FILENAMEONLY];
                NAME_MATCHES = FILELISTING_TRACKER[SOURCEFILE,FILELISTING_INFO[FILEIDX,j,"CHECKSUM"],FILEPATHONLY];

                # we need to check that MATCH tree doesnt contain another copy of the said file too.
                LOCAL_PATH_MATCHES = FILELISTING_TRACKER[j,FILELISTING_INFO[FILEIDX,j,"CHECKSUM"],FILENAMEONLY];
                LOCAL_NAME_MATCHES = FILELISTING_TRACKER[j,FILELISTING_INFO[FILEIDX,j,"CHECKSUM"],FILEPATHONLY];

                MATCHED=0;

                # Detect file rename for added file
                # To detect that file has has been renamed to result in  MATCH, we need to find a source file matching the conditions:
                # 1. checksum matches
                # 2. a) filename is either a substring or a superstring, excluding extension, 
                #    b) but not matching exacatly
                # 3. exists in SOURCE in the same directory
                # 4. does not exist in MATCH in the place it exists in SOURCE
                # 5. only one such file satisfying the above exists
                if ((MATCHED==0) && (""!=NAME_MATCHES) && ("\0" FILENAMEONLY !=NAME_MATCHES)) {
                    MATCHED=0;

                    # conditions (1) and (3) matched

                    MATCHCOUNT = split(NAME_MATCHES, MATCHINGFILES, "\0");

                    for (m=1; m<=length(MATCHINGFILES); m++) {
                        if (""==MATCHINGFILES[m]) {
                            continue;
                        }

                        CANDIDATE_ENTRY = MATCHINGFILES[m];
                        CANDIDATE_TOKEN=gensub("\\..*","","",MATCHINGFILES[m]);
                        FILENAME_TOKEN=gensub("\\..*","","",FILENAMEONLY);

                        COPIES = "";

                        if ((""!=CANDIDATE_TOKEN)&&(CANDIDATE_ENTRY!=FILENAMEONLY)&&((0!=index(CANDIDATE_TOKEN, FILENAME_TOKEN)) || (0!=index(FILENAME_TOKEN, CANDIDATE_TOKEN)))) {
                            
                            # condition (2) matched
                            # now we have suspects which are likely related to the file, which we interprete as copies.

                            # set "c" for copy
                            TRACK[j] = 99;

#print "r1" CANDIDATE_ENTRY
                            nameof_sourcefile = gensub("[/]", "\0", "g",gensub("^[^/]*[\\/]", "./", "",gensub("\\\\", "/", "g", FILEPATHONLY CANDIDATE_ENTRY)));
                            sizeof_matchfile = FILELISTING_INFO[FILELISTING_INDEX[nameof_sourcefile],j,"SIZE"]

                            if (""==sizeof_matchfile) {
                                #condition (4) satisfied
                                
                                if (MATCHED==0) {

                                        # RENAME
                                        # assume condition (5) has been satisfied and treat as a rename
                                        # set "r" for renamed
                                        TRACK[j] = 114;
                                        MATCHED = 1;
                                        
                                        REFERENCETARGET[j]="\0(" SRC ")" CANDIDATE_ENTRY

                                } else {
                                        
                                        # condition (5) broken
                                        # downgrade to "c" for copy

                                        TRACK[j] = 99;
                                }
                            }
                            COPIES = COPIES "\0(" SRC ")" CANDIDATE_ENTRY
                        }

                    }
                    if (TRACK[j]==99) {
                       REFERENCETARGET[j] = COPIES
                    }
                } 

                # Detect file move for added file
                # To detect that file has has been moved from somewhere from SOURCE, the conditions are:
                # 1. filename and checksum match 
                # 2. exists in a different directory in SOURCE than in MATCH
                # 3. and only one such file exists
                # 4. MATCH did not already contain the file in (2)
                if ((MATCHED==0) && (""!=PATH_MATCHES) && ("\0" FILEPATHONLY !=PATH_MATCHES)) {
                    
                    MATCHED = 0;

                    # condition (1) passed
                    # no we know that copios of the file exist somewhere

                    # set "c" for copy
                    TRACK[j] = 99;

                    MATCHCOUNT = split(PATH_MATCHES, MATCHINGFILES, "\0");

                    for (m=1; m<=length(MATCHINGFILES); m++) {
                        if (""==MATCHINGFILES[m]) {
                            continue;
                        }

                        CANDIDATE_ENTRY = MATCHINGFILES[m];

                        # Since every entry starts with \0 there would be one extra entry at the beginning of the array
                        if (FILEPATHONLY!=CANDIDATE_ENTRY) {
                            # conditions (2) and (3) passed

#print "m1" CANDIDATE_ENTRY
                            nameof_sourcefile = gensub("[/]", "\0", "g",gensub("^[^/]*[\\/]", "./", "",gensub("\\\\", "/", "g", CANDIDATE_ENTRY FILENAMEONLY)));
                            sizeof_matchfile =  FILELISTING_INFO[FILELISTING_INDEX[nameof_sourcefile],j,"SIZE"];

                            if (""==sizeof_matchfile) {
                                # condition (4) passed
                                
                                if (MATCHED==0) {

                                    # MOVE detected
                                    # assume condition (3) passed
                                    # set "m" for moved

                                    TRACK[j] = 109;
                                    MATCHED=1;

                                    REFERENCETARGET[j]="\0(" TGT ")" CANDIDATE_ENTRY

                                } else {

                                    # condition (3) broken
                                    # downgread to "c" for copy

                                    TRACK[j] = 99;
                                }

                            }
                        } 
                    }
                    if (TRACK[j]==99) {
                        filenamepos = index(PATH_MATCHES "\0", "\0" FILEPATHONLY "\0");
                        if (filenamepos==0) {
                            REFERENCETARGET[j]="\0(" SRC ")" substr(PATH_MATCHES,2);
                        } else {
                            REFERENCETARGET[j]="\0(" SRC ")" substr(substr(PATH_MATCHES, 1, filenamepos-1) substr(PATH_MATCHES, filenamepos+length(FILEPATHONLY)+1),2);
                        }
                    }
                    
                } 
            }
            FILELISTING_INFO[FILEIDX,0,"CODE"] = FILELISTING_INFO[FILEIDX,0,"CODE"] sprintf("%c",0+CODE[j]);
            FILELISTING_INFO[FILEIDX,0,"TRACK"] = FILELISTING_INFO[FILEIDX,0,"TRACK"] sprintf("%c",0+TRACK[j]);
        }

        asort(CODE, CODE_SORT);
        prevcode=0;
        for (j=1; j<=length(CODE); j++) {
            if (prevcode != CODE[j]) {
                if (""!=REFERENCETARGET[j]) {
                    TGT = "S" j
                    if (FILECOUNT==2) {
                        if (j==1) { TGT="SOURCE" }
                        if (j==2) { TGT="MATCH" }
                    }
                    if (FILECOUNT==3) {
                        if (j==1) { TGT="M1" }
                        if (j==2) { TGT="SOURCE" }
                        if (j==3) { TGT="M2" }
                    }
                    ISNOTDEL = ("0"!=sprintf("%c",0+CODE[j]))
                    OPCODE = sprintf("%c",0+TRACK[j])
                    MSG = gensub(" *$", "", "", MESSAGE[MESSAGE_LANG,OPCODE ISNOTDEL])
                    FILELISTING_INFO[FILEIDX,0,"REFERENCETARGET"] = FILELISTING_INFO[FILEIDX,0,"REFERENCETARGET"] "[" TGT "" MSG "{" substr(gensub("\0", "; ", "g", REFERENCETARGET[j]),3) "}] ";
                }
            }
            prevcode = CODE[j];
        }
        BUFFER = BUFFER FILELISTING_INFO[FILEIDX,0,"CODE"] SEP MESSAGE[MESSAGE_LANG,FILELISTING_INFO[FILEIDX,0,"CODE"]] SEP FILELISTING_INFO[FILEIDX,0,"TRACK"] SEP FILELISTING_INFO[FILEIDX,0,"REFERENCETARGET"];
        print BUFFER;
    }
}

function julianday (datestring)
{
        match(datestring, "([0-9][0-9][0-9][0-9])-([0-9][0-9])-([0-9][0-9]).*", COMPS)
        val_y = COMPS[1]
        val_m = COMPS[2]
        val_d = COMPS[3]

        if (int(val_m)<3) {
                val_m += 12;
                val_y -= 1;
        }
        val_C = int(val_y/100);
        val_X = val_y % 100;

        val_J0 = 1721119.5;
        val_J1 = int(146097*val_C/4);
        val_J2 = int(36525*val_X/100);
        val_J3 = int((153* val_m - 457)/5);
        val_j = val_J1 + val_J2 + val_J3 + val_d - 1 + val_J0;
        return sprintf("%0.5f",val_j)

}

function calendarday (val_j)
{
        val_X2 = val_j - 1721119.5;
        val_C2 = int((8*val_X2 + 7)/292194);
        val_X1 = val_X2 - int(146097*val_C2/4);
        val_C1 = int((200*val_X1 + 199)/73050);
        val_X0 = val_X1 - int(36525*val_C1/100);
        val_y = 100*val_C2 + val_C1;
        val_m = int((10*val_X0 + 923)/306);
        val_d = val_X0 - int((153* val_m - 457)/5) + 1;

        if (val_m>12) {
                val_m -= 12;
                val_y += 1;
        }
        return sprintf("%02d-%02d-%02d", val_y, val_m, val_d);
}
